@@ -2,17 +2,17 @@
 store the current version info of the server.
 
 """
+
 import re
-from typing import List
 
 # Version string must appear intact for automatic versioning
-__version__ = "2.2.1"
+__version__ = "2.16.0"
 
 # Build up version_info tuple for backwards compatibility
 pattern = r"(?P<major>\d+).(?P<minor>\d+).(?P<patch>\d+)(?P<rest>.*)"
 match = re.match(pattern, __version__)
 assert match is not None
-parts: List[object] = [int(match[part]) for part in ["major", "minor", "patch"]]
+parts: list[object] = [int(match[part]) for part in ["major", "minor", "patch"]]
 if match["rest"]:
     parts.append(match["rest"])
 version_info = tuple(parts)
